"""
Written by Newell Moser
NRC Postdoctoral Fellow
National Institute of Standards and Technology
December 2, 2024

Tests for calcualating the minimum rotated bounding box of a point cloud.
"""

import os
import numpy as np
import imppy3d.bounding_box as box


from matplotlib import pyplot as plt
# Set constants related to plotting (for MatPlotLib)
SMALL_SIZE = 10
MEDIUM_SIZE = 12
BIGGER_SIZE = 14

plt.rc('font', size=MEDIUM_SIZE)         # Controls default text sizes
plt.rc('axes', titlesize=BIGGER_SIZE)    # Fontsize of the axes title
plt.rc('axes', labelsize=MEDIUM_SIZE)    # Fontsize of the x and y labels
plt.rc('xtick', labelsize=MEDIUM_SIZE)   # Fontsize of the tick labels
plt.rc('ytick', labelsize=MEDIUM_SIZE)   # Fontsize of the tick labels
plt.rc('legend', fontsize=MEDIUM_SIZE)   # Legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # Fontsize of the figure title


def connect_box_corners(corner_pnts):
    """
    Helper function that connects all of the corners of a 3D bounding
    box into a single, continuous line for easier plotting. The order
    of the corners is based on the BBox class defined in
    volume_image_processing.py
    """
    return np.array([
        corner_pnts[0,:],
        corner_pnts[1,:],
        corner_pnts[2,:],
        corner_pnts[3,:],
        corner_pnts[0,:],
        corner_pnts[4,:],
        corner_pnts[5,:],
        corner_pnts[1,:],
        corner_pnts[5,:],
        corner_pnts[6,:],
        corner_pnts[2,:],
        corner_pnts[6,:],
        corner_pnts[7,:],
        corner_pnts[3,:],
        corner_pnts[7,:],
        corner_pnts[4,:] ])


def farthest_dist_between_pnts(pnt_arr):
    num_pnt = pnt_arr.shape[0]
    n_dim = pnt_arr.shape[1]
    dist_arr = np.zeros((num_pnt*num_pnt, n_dim))

    dist_i = 0
    for m, pnt1 in enumerate(pnt_arr):
        for n in np.arange(start=m, stop=num_pnt):
            pnt2 = pnt_arr[n]
            dist_arr[dist_i] = np.sqrt(np.sum(np.power(pnt1 - pnt2, 2)))
            dist_i += 1

    return np.amax(dist_arr)


def check_if_pnt_in_arr(pnt_0, pnt_arr, TOL):
    num_pnt = pnt_arr.shape[0]
    n_dim = pnt_arr.shape[1]

    TOL2 = TOL*TOL

    for cur_pnt in pnt_arr:
        cur_dist = np.sum(np.power(pnt_0 - cur_pnt, 2))

        if cur_dist <= TOL2:
            return True

    return False


def create_dummy_pnts(L=16.0, W=8.0, T=4.0):
    # [X, Y, Z]
    pnt1 = np.array([0.0, 0.0, 0.0])
    pnt2 = np.array([L, 0.0, 0.0])
    pnt3 = np.array([L, W, 0.0])
    pnt4 = np.array([0.0, W, 0.0])
    pnt5 = np.array([0.0, 0.0, T])
    pnt6 = np.array([L, 0.0, T])
    pnt7 = np.array([L, W, T])
    pnt8 = np.array([0.0, W, T])
    #pnt9 = np.array([L/2, W/2, T/2])

    pnt_arr = np.vstack((pnt1, pnt2, pnt3, pnt4, pnt5, pnt6, pnt7, pnt8))

    z_deg = 45.0
    y_deg = 45.0
    x_deg = 45.0

    x_rad = np.deg2rad(x_deg)
    y_rad = np.deg2rad(y_deg)
    z_rad = np.deg2rad(z_deg)

    cx = np.cos(x_rad)
    sx = np.sin(x_rad)
    R_x = np.array([ 
        [1., 0., 0.],
        [0., cx, -sx],
        [0., sx, cx] ])

    cy = np.cos(y_rad)
    sy = np.sin(y_rad)
    R_y = np.array([
        [cy, 0., sy],
        [0., 1., 0.],
        [-sy, 0., cy] ])

    cz = np.cos(z_rad)
    sz = np.sin(z_rad)
    R_z = np.array([
        [cz, -sz, 0.],
        [sz, cz, 0.],
        [0., 0., 1.] ])

    R_zyx = np.matmul(R_z, np.matmul(R_y, R_x))

    # Finally, apply random 3D rotation
    pnt_arr = np.dot(R_zyx, np.transpose(pnt_arr))
    pnt_arr = np.transpose(pnt_arr)

    return pnt_arr


def bbox_svd():
    test_name = "\nTEST: BOUNDING BOX FIT USING SVD SEARCH..."
    print(test_name)

    pnts_src = create_dummy_pnts()
    pnts_bool_arr = np.full(pnts_src.shape[0], False, dtype=bool)
    TOL = 0.05*farthest_dist_between_pnts(pnts_src)

    try:
        bbox_obj = box.min_bounding_box(pnts_src, calc_hull=False, search=3)
    except:
        print(f"\nERROR: Failed to fit a bounding box using SVD search.")
        return [1, test_name + " ERROR"]

    bbox_corners = bbox_obj.corners

    for m, cur_pnt_src in enumerate(pnts_src):
        temp_bool = check_if_pnt_in_arr(cur_pnt_src, bbox_corners, TOL)
        pnts_bool_arr[m] = temp_bool

    if not all(pnts_bool_arr):
        print(f"\nERROR: Bounding box not within tolerance of actual points.")
        return [1, test_name + " ERROR"]

    return [0, test_name + " SUCCESS"] # (False) No errors


def bbox_exhaustive():
    test_name = "\nTEST: BOUNDING BOX FIT USING EXHAUSTIVE SEARCH..."
    print(test_name)

    pnts_src = create_dummy_pnts()
    pnts_bool_arr = np.full(pnts_src.shape[0], False, dtype=bool)
    TOL = 0.05*farthest_dist_between_pnts(pnts_src)

    try:
        bbox_obj = box.min_bounding_box(pnts_src, calc_hull=False, search=0)
    except:
        print(f"\nERROR: Failed to fit a bounding box using exhaustive search.")
        return [1, test_name + " ERROR"]

    bbox_corners = bbox_obj.corners

    for m, cur_pnt_src in enumerate(pnts_src):
        temp_bool = check_if_pnt_in_arr(cur_pnt_src, bbox_corners, TOL)
        pnts_bool_arr[m] = temp_bool

    if not all(pnts_bool_arr):
        print(f"\nERROR: Bounding box not within tolerance of actual points.")
        return [1, test_name + " ERROR"]

    return [0, test_name + " SUCCESS"] # (False) No errors


def bbox_LWT():
    test_name = "\nTEST: BOUNDING BOX FIT USING LWT SEARCH..."
    print(test_name)

    pnts_src = create_dummy_pnts(L=512, W=8, T=4)
    pnts_bool_arr = np.full(pnts_src.shape[0], False, dtype=bool)
    TOL = 0.05*farthest_dist_between_pnts(pnts_src)

    try:
        bbox_obj = box.min_bounding_box(pnts_src, calc_hull=False, search=4)
    except:
        print(f"\nERROR: Failed to fit a bounding box using LWT search.")
        return [1, test_name + " ERROR"]

    bbox_corners = bbox_obj.corners

    print(f"\npnts_src = \n{pnts_src}")
    print(f"\nbbox_corners = \n{bbox_corners}")

    for m, cur_pnt_src in enumerate(pnts_src):
        temp_bool = check_if_pnt_in_arr(cur_pnt_src, bbox_corners, TOL)
        pnts_bool_arr[m] = temp_bool

    if not all(pnts_bool_arr):
        print(f"\nERROR: Bounding box not within tolerance of actual points.")
        return [1, test_name + " ERROR"]

    return [0, test_name + " SUCCESS"] # (False) No errors


if __name__ == '__main__':

    flag_300, msg_300 = bbox_svd()
    flag_301, msg_301 = bbox_exhaustive()
    flag_302, msg_302 = bbox_LWT()

    print(f"\n\n\n---------- SUMMARY ----------")
    print(msg_300)
    print(msg_301)
    print(msg_302)
    print("\n")


"""
Written by Newell Moser
NRC Postdoctoral Fellow
National Institute of Standards and Technology
December 2, 2024

Test import/export of image stacks.
"""

import os, pathlib, glob
import numpy as np
from skimage.util import img_as_ubyte, img_as_float32
import imppy3d.import_export as imex


def del_all_img_files(dir_path='./'):
    file_ext = (".tif", ".tiff", ".png", ".jpg", ".jpeg", ".bmp", ".pgm")
    img_names = []

    # All files in directory
    files = glob.glob(dir_path + '*') 
    if files:
        for cur_name in files:
            # Makes forward slashes into backward slashes for W10, and
            # also makes the directory string all lowercase
            cur_name = os.path.normcase(cur_name)
            # Only keep image files
            if (cur_name.lower()).endswith(file_ext):
                img_names.append(cur_name)

    # If empty, nothing to delete
    if not img_names:
        return 

    img_names = list(dict.fromkeys(img_names)) # Removes duplicates
    img_names.sort() # Ascending order based on the string file names

    # Deletes all of the image files found in this directory
    for cur_name in img_names:
        file_to_rem = pathlib.Path(cur_name)
        file_to_rem.unlink()


def io_img_stack_8bit(dir_path='./'):
    test_name = "\nTEST: IMPORTING/EXPORTING UNSIGNED 8-BIT IMAGE STACKS..."
    print(test_name)

    os.makedirs(dir_path, exist_ok=True)

    # Create an empty image stack containing 4 images
    imgs_out_8bit = np.zeros((4,128,128), dtype=np.uint8)
    imgs_out_8bit[0,0,0] = 250
    imgs_out_substr = "test_imgs_8bit.tif"

    try:
        imex.save_image_seq(imgs_out_8bit, dir_path, imgs_out_substr)
    except:
        print(f"\nERROR: Failed to save 8-bit image stack.")
        return [1, test_name + " ERROR"] # (True) Errors occurred

    try:
        imgs_in_8bit, imgs_names = imex.load_image_seq(dir_path, 
            file_name_in=imgs_out_substr[0:-4], indices_in=(4,),
            img_bitdepth_in='uint8')
    except:
        print(f"\nERROR: Failed to load 8-bit image stack.")
        return [1, test_name + " ERROR"] # (True) Errors occurred

    if imgs_out_8bit.dtype != imgs_in_8bit.dtype:
        print(f"\nERROR: The data type of the imported image stack is not the")
        print(f"same as the exported image stack.")
        return [1, test_name + " ERROR"]

    if imgs_out_8bit.shape != imgs_in_8bit.shape:
        print(f"\nERROR: The shape of the imported image stack is not the")
        print(f"same as the exported image stack.")
        return [1, test_name + " ERROR"]

    TOL = 1.0E-2
    imgs_diff = np.absolute(img_as_float32(imgs_out_8bit) - img_as_float32(imgs_in_8bit))
    max_diff = np.amax(imgs_diff)

    if max_diff >= TOL:
        print(f"\nERROR: Imported image stack differs from the exported image"\
            " stack.")
        return [1, test_name + " ERROR"]

    del_all_img_files(dir_path)
    return [0, test_name + " SUCCESS"] # (False) No errors


def io_img_stack_16bit(dir_path='./'):
    test_name = "\nTEST: IMPORTING/EXPORTING UNSIGNED 16-BIT IMAGE STACKS..."
    print(test_name)

    os.makedirs(dir_path, exist_ok=True)

    # Create an empty image stack containing 4 images
    imgs_out_16bit = np.zeros((4,128,128), dtype=np.uint16)
    imgs_out_16bit[0,0,0] = 65000
    imgs_out_substr = "test_imgs_16bit.tif"

    try:
        imex.save_image_seq(imgs_out_16bit, dir_path, imgs_out_substr)
    except:
        print(f"\nERROR: Failed to save 16-bit image stack.")
        return [1, test_name + " ERROR"] # (True) Errors occurred

    try:
        imgs_in_16bit, imgs_names = imex.load_image_seq(dir_path, 
            file_name_in=imgs_out_substr[0:-4], indices_in=(4,), 
            img_bitdepth_in='uint16')
    except:
        print(f"\nERROR: Failed to load 16-bit image stack.")
        return [1, test_name + " ERROR"] # (True) Errors occurred

    if imgs_out_16bit.dtype != imgs_in_16bit.dtype:
        print(f"\nERROR: The data type of the imported image stack is not the")
        print(f"same as the exported image stack.")
        return [1, test_name + " ERROR"]

    if imgs_out_16bit.shape != imgs_in_16bit.shape:
        print(f"\nERROR: The shape of the imported image stack is not the")
        print(f"same as the exported image stack.")
        return [1, test_name + " ERROR"]

    TOL = 1.0E-2
    imgs_diff = np.absolute(img_as_float32(imgs_out_16bit) - img_as_float32(imgs_in_16bit))
    max_diff = np.amax(imgs_diff)

    if max_diff >= TOL:
        print(f"\nERROR: Imported image stack differs from the exported image"\
            " stack.")
        return [1, test_name + " ERROR"]

    del_all_img_files(dir_path)
    return [0, test_name + " SUCCESS"] # (False) No errors


def io_multipage_tiff_stack(dir_path='./'):
    test_name = "\nTEST: IMPORTING/EXPORTING MULTI-PAGE TIFF STACKS..."
    print(test_name)

    os.makedirs(dir_path, exist_ok=True)

    # Create an empty image stack containing 4 images
    imgs_out_8bit = np.zeros((8,128,128), dtype=np.uint8)
    imgs_out_8bit[0,0,0] = 250
    imgs_path = os.path.normcase(dir_path + "test_imgs_8bit.tiff")

    try:
        imex.save_multipage_image(imgs_out_8bit, imgs_path)
    except:
        print(f"\nERROR: Failed to save 8-bit multi-page tiff stack.")
        return [1, test_name + " ERROR"] # (True) Errors occurred

    try:
        imgs_in_8bit, imgs_names = imex.load_multipage_image(imgs_path, 
            indices_in=(8,), img_bitdepth_in='uint8')
    except:
        print(f"\nERROR: Failed to load 8-bit multi-page tiff stack.")
        return [1, test_name + " ERROR"] # (True) Errors occurred

    if imgs_out_8bit.dtype != imgs_in_8bit.dtype:
        print(f"\nERROR: The data type of the imported image stack is not the")
        print(f"same as the exported image stack.")
        return [1, test_name + " ERROR"]

    if imgs_out_8bit.shape != imgs_in_8bit.shape:
        print(f"\nERROR: The shape of the imported image stack is not the")
        print(f"same as the exported image stack.")
        return [1, test_name + " ERROR"]

    TOL = 1.0E-2
    imgs_diff = np.absolute(img_as_float32(imgs_out_8bit) - img_as_float32(imgs_in_8bit))
    max_diff = np.amax(imgs_diff)

    if max_diff >= TOL:
        print(f"\nERROR: Imported image stack differs from the exported image"\
            " stack.")
        return [1, test_name + " ERROR"]

    del_all_img_files(dir_path)
    return [0, test_name + " SUCCESS"] # (False) No errors


if __name__ == '__main__':

    temp_dir_path = "./local_swap/"
    flag_000, msg_000 = io_img_stack_8bit(temp_dir_path)
    flag_001, msg_001 = io_img_stack_16bit(temp_dir_path)
    flag_002, msg_002 = io_multipage_tiff_stack(temp_dir_path)

    print(f"\n\n\n---------- SUMMARY ----------")
    print(msg_000)
    print(msg_001)
    print(msg_002)
    print("\n")


"""
Written by Newell Moser
NRC Postdoctoral Fellow
National Institute of Standards and Technology
December 2, 2024

Tests for calcualating the minimum rotated bounding box of a point cloud.
"""

import os
import numpy as np
import imppy3d.ski_driver_functions as sdrv


def make_dummy_img_stack(vec_dir='z'):
    imgs = np.zeros((129,129,129), dtype=np.uint8)

    rad_maj = 31
    rad_min = 15
    mid_i = int(np.floor(129/2.0))
    
    if vec_dir=='z': # Along image coordinates
        imgs[mid_i-rad_maj:mid_i+rad_maj+1, mid_i-rad_min:mid_i+rad_min+1,
            mid_i-rad_min:mid_i+rad_min+1] = 255

    elif vec_dir=='y': # Along row coordinates
        imgs[mid_i-rad_min:mid_i+rad_min+1, mid_i-rad_maj:mid_i+rad_maj+1, 
            mid_i-rad_min:mid_i+rad_min+1] = 255

    elif vec_dir=='x': # Along column coordinates
        imgs[mid_i-rad_min:mid_i+rad_min+1, mid_i-rad_min:mid_i+rad_min+1,
            mid_i-rad_maj:mid_i+rad_maj+1] = 255

    return imgs


def make_dummy_img():
    img_width = 129

    img = np.zeros((img_width,img_width), dtype=np.uint8)
    mid_i = int(np.floor(img_width/2.0))

    rad_maj = 31
    rad_min = 31
    img[mid_i-rad_maj:mid_i+rad_maj+1, mid_i-rad_min:mid_i+rad_min+1] = 127

    rad_maj = 15
    rad_min = 15
    img[mid_i-rad_maj:mid_i+rad_maj+1, mid_i-rad_min:mid_i+rad_min+1] = 255

    return img


def run_scikit_filters():
    # These are wrappers to the Scikit-Image library. Hence, no need to test
    # the actual accuracy of the filters since that is a verification already
    # handled by the Scikit-Image library.
    test_name = "\nTEST: EXECUTING SCIKIT-IMAGE WRAPPERS..."
    print(test_name)

    img_gray = make_dummy_img()
    img_bin = img_gray.copy()
    img_bin[img_gray > 0] = 255

    try:
        sharp_radius = 2    # int
        sharp_amount = 0.2  # float
        params = ["unsharp_mask", int(sharp_radius), sharp_amount] 
        img_temp = sdrv.apply_driver_sharpen(img_gray, params, quiet_in=True)
    except:
        print(f"\nERROR: Failed to apply unsharp filter.")
        return [1, test_name + " ERROR"]

    try:
        global_thresh = 200
        params = ["global_threshold", global_thresh]
        img_temp = sdrv.apply_driver_thresholding(img_gray, params, quiet_in=True)
    except:
        print(f"\nERROR: Failed to apply global thresholding filter.")
        return [1, test_name + " ERROR"]

    try:
        h_out = 0.01
        patch_size_out = 3
        patch_dist_out = 5
        params = ["nl_means", h_out, patch_size_out, patch_dist_out]
        img_temp = sdrv.apply_driver_denoise(img_gray, params, quiet_in=True)
    except:
        print(f"\nERROR: Failed to apply nonlocal-means denoising filter.")
        return [1, test_name + " ERROR"]

    return [0, test_name + " SUCCESS"] # (False) No errors

    try:
        operation_type_out = 2
        footprint_type_out = 0
        n_radius_out = 1
        params = [operation_type_out, footprint_type_out, n_radius_out]
        img_temp = sdrv.apply_driver_morph(img_bin, params, quiet_in=True)
    except:
        print(f"\nERROR: Failed to apply binary morphological filter.")
        return [1, test_name + " ERROR"]

    return [0, test_name + " SUCCESS"] # (False) No errors
    

if __name__ == '__main__':

    flag_400, msg_400 = run_scikit_filters()

    print(f"\n\n\n---------- SUMMARY ----------")
    print(msg_400)
    print("\n")


"""
Written by Newell Moser
NRC Postdoctoral Fellow
National Institute of Standards and Technology
December 2, 2024

Test creation of (3D) VTK models, which can be used by ParaView.
"""

import os, pathlib, glob
import numpy as np
from skimage import draw
from skimage.util import img_as_ubyte, img_as_float32
import imppy3d.vtk_api as vapi


def del_all_vtk_files(dir_path='./'):
    file_ext = (".vtk", ".vtu")
    img_names = []

    # All files in directory
    files = glob.glob(dir_path + '*') 
    if files:
        for cur_name in files:
            # Makes forward slashes into backward slashes for W10, and
            # also makes the directory string all lowercase
            cur_name = os.path.normcase(cur_name)
            # Only keep image files
            if (cur_name.lower()).endswith(file_ext):
                img_names.append(cur_name)

    # If empty, nothing to delete
    if not img_names:
        return 

    img_names = list(dict.fromkeys(img_names)) # Removes duplicates
    img_names.sort() # Ascending order based on the string file names

    # Deletes all of the image files found in this directory
    for cur_name in img_names:
        file_to_rem = pathlib.Path(cur_name)
        file_to_rem.unlink()


def make_dummy_ellipsoid_img_stack():
    # Draw an ellipsoid
    # ********** USER INPUTS ********** 
    # x^2/a^2 + y^2/b^2 + z^2/c^2 = 1
    # Define the semi-major and semi-minor (radii) axes.
    ellip_a = 15 # Aligned to Z-axis (Image Index)
    ellip_b = 31 # Aligned to Y-axis (Row Index)
    ellip_c = 63 # Aligned to X-axis (Column Index)
    # ********************************* 

    img_ellip = draw.ellipsoid(ellip_a, ellip_b, ellip_c)
    img_ellip = img_as_ubyte(img_ellip)

    return img_ellip


def make_dummy_box_img_stack(vec_dir='z'):
    imgs = np.zeros((129,129,129), dtype=np.uint8)

    rad_maj = 31
    rad_min = 31
    mid_i = int(np.floor(129/2.0))
    
    if vec_dir=='z': # Along image coordinates
        imgs[mid_i-rad_maj:mid_i+rad_maj+1, mid_i-rad_min:mid_i+rad_min+1,
            mid_i-rad_min:mid_i+rad_min+1] = 255

    elif vec_dir=='y': # Along row coordinates
        imgs[mid_i-rad_min:mid_i+rad_min+1, mid_i-rad_maj:mid_i+rad_maj+1, 
            mid_i-rad_min:mid_i+rad_min+1] = 255

    elif vec_dir=='x': # Along column coordinates
        imgs[mid_i-rad_min:mid_i+rad_min+1, mid_i-rad_min:mid_i+rad_min+1,
            mid_i-rad_maj:mid_i+rad_maj+1] = 255

    return imgs


def create_vtk_polydata_surf():
    # Creates a surface from a voxel model of an ellipsoid
    test_name = "\nTEST: CREATE VTK POLYDATA SURFACE USING MARCHING CUBES..."
    print(test_name)

    imgs_b = make_dummy_box_img_stack()
    n_voxels = 63*63*63

    try:
        verts, faces, normals, vals = vapi.convert_voxels_to_surface(imgs_b)
    except:
        print(f"\nERROR: Failed to apply marching cubes to the voxel model.")
        return [1, test_name + " ERROR"] # (True) Errors occurred

    try:
        vtk_obj = vapi.make_vtk_surf_mesh(verts, faces, vals, smth_iter=0)
    except:
        print(f"\nERROR: Failed to create a VTK polydata object.")
        return [1, test_name + " ERROR"] # (True) Errors occurred

    TOL = 0.05*n_voxels
    vol_diff = np.absolute(vtk_obj.volume - n_voxels)

    if vol_diff >= TOL:
        print(f"\nERROR: Created surface model does not contain the correct"\
            " volume.")
        return [1, test_name + " ERROR"] # (True) Errors occurred

    return [0, test_name + " SUCCESS"] # No errors


def create_vtk_unstructured_grid():
    # Creates a voxel model of an ellipsoid
    test_name = "\nTEST: CREATE VTK UNSTRUCTURED VOXEL MODEL (C EXTENSION)..."
    print(test_name)

    imgs_b = make_dummy_box_img_stack() # 63 x 63 x 63 cuboid
    n_voxels = 63*63*63

    try:
        vtk_obj = vapi.make_vtk_unstruct_grid(imgs_b, quiet_in=True)
    except:
        print(f"\nERROR: Failed to create VTK voxel model.")
        return [1, test_name + " ERROR"] # (True) Errors occurred

    # Check that the correct number of voxels were created
    if vtk_obj.n_cells != n_voxels:
        print(f"\nERROR: Created voxel model does not contain the correct"\
            " number of cells.")
        return [1, test_name + " ERROR"] # (True) Errors occurred

    return [0, test_name + " SUCCESS"] # No errors


def create_vtk_uniform_grid():
    # Creates a VTK image data model of an ellipsoid
    test_name = "\nTEST: CREATE VTK IMAGE DATA MODEL..."
    print(test_name)

    imgs_b = make_dummy_box_img_stack()
    n_voxels = 63*63*63

    try:
        vtk_obj = vapi.make_vtk_uniform_grid(imgs_b)
    except:
        print(f"\nERROR: Failed to create VTK image data model.")
        return [1, test_name + " ERROR"] # (True) Errors occurred

    vtk_arr = vtk_obj.cell_data['values']
    vtk_arr2 = vtk_arr[vtk_arr > 128]

    if vtk_arr2.size != n_voxels:
        print(f"\nERROR: Created uniform grid does not contain the correct"\
            " number of cells.")
        return [1, test_name + " ERROR"] # (True) Errors occurred

    return [0, test_name + " SUCCESS"] # No errors


if __name__ == '__main__':

    flag_100, msg_100 = create_vtk_unstructured_grid()
    flag_101, msg_101 = create_vtk_polydata_surf()
    flag_102, msg_102 = create_vtk_uniform_grid()

    print(f"\n\n\n---------- SUMMARY ----------")
    print(msg_100)
    print(msg_101)
    print(msg_102)
    print("\n")


import numpy as np

# "cimport" is used to import special compile-time information
# about the numpy module (this is stored in a file numpy.pxd which is
# currently part of the Cython distribution).
cimport numpy as np

cimport cython

# We now need to fix a datatype for our arrays. I've used the variable
# DTYPE1 for this, which is assigned to the usual NumPy runtime
# type info object.
DTYPE1 = np.uint8
DTYPE2 = np.uint32
DTYPE3 = np.float32
DTYPE4 = np.int32
DTYPE5 = np.uint64

# "ctypedef" assigns a corresponding compile-time type to DTYPE1_t. For
# every type in the numpy module there's a corresponding compile-time
# type with a _t-suffix.
# https://numpy.org/doc/1.19/user/basics.types.html?highlight=dtypes
ctypedef np.uint8_t DTYPE1_t
ctypedef np.uint32_t DTYPE2_t
ctypedef np.float32_t DTYPE3_t
ctypedef np.int32_t DTYPE4_t
ctypedef np.uint64_t DTYPE5_t

@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function

# "def" can type its arguments but not have a return type. The type of the
# arguments for a "def" function is checked at run-time when entering the
# function
def get_voxel_info(np.ndarray[DTYPE1_t, ndim=3] img_arr, DTYPE4_t debug_flag):

    # Some basic checks for the input arrays
    cdef DTYPE2_t img_dim = img_arr.ndim
    if img_dim != 3:
        print(f"Number of dimensions in img_arr: {img_dim}")
        raise ValueError("img_arr must be a 3D Numpy array")
    assert img_arr.dtype == DTYPE1

    # Constants to be used later
    cdef DTYPE1_t ONE_UINT8 = 1
    cdef DTYPE2_t ONE_UINT32 = 1
    cdef Py_ssize_t ONE_PYSZT = 1
    cdef DTYPE3_t HALF = 0.5
    cdef DTYPE2_t MAX_UINT32 = 4294967295
    cdef DTYPE2_t EIGHT = 8

    # Py_ssize_t is essentially just an int (for the Python purist) which
    # should be used for index in Cython. Can actually use an int here 
    # and it would be fine
    cdef DTYPE2_t n_imgs = img_arr.shape[0]
    cdef DTYPE2_t n_rows = img_arr.shape[1]
    cdef DTYPE2_t n_cols = img_arr.shape[2]
    cdef Py_ssize_t ii, rr, cc, xc, yc, zc # For-loop indices

    # The number of voxels that will be created. Initialize the Numpy
    # connectivity array that will be used to create the VTK model
    cdef DTYPE2_t num_cells = 0
    cdef DTYPE2_t num_verts = 0
    cdef DTYPE1_t temp_UINT8

    if debug_flag > 0:
        print(f"\nTemporarily allocating "\
            f"{(n_imgs+1)*(n_rows+1)*(n_cols+1)*4/1E6} MB of memory")

    cdef np.ndarray[DTYPE2_t, ndim=3] vert_ids = np.ones(
        (n_imgs+1, n_rows+1, n_cols+1), dtype=DTYPE2)

    cdef np.ndarray[DTYPE3_t, ndim=1] nd1 = np.zeros(3, dtype=DTYPE3)
    cdef np.ndarray[DTYPE3_t, ndim=1] nd2 = np.zeros(3, dtype=DTYPE3)
    cdef np.ndarray[DTYPE3_t, ndim=1] nd3 = np.zeros(3, dtype=DTYPE3)
    cdef np.ndarray[DTYPE3_t, ndim=1] nd4 = np.zeros(3, dtype=DTYPE3)
    cdef np.ndarray[DTYPE3_t, ndim=1] nd5 = np.zeros(3, dtype=DTYPE3)
    cdef np.ndarray[DTYPE3_t, ndim=1] nd6 = np.zeros(3, dtype=DTYPE3)
    cdef np.ndarray[DTYPE3_t, ndim=1] nd7 = np.zeros(3, dtype=DTYPE3)
    cdef np.ndarray[DTYPE3_t, ndim=1] nd8 = np.zeros(3, dtype=DTYPE3)

    cdef DTYPE2_t nd1_id, nd2_id, nd3_id, nd4_id, nd5_id, nd6_id, nd7_id, nd8_id

    # Initialize the vert_ids 3D array to the largest possible value that
    # can be stored in a UINT32. Any vertex with a corresponding value in
    # this array equal to this maximum value will not be used in the final
    # voxel model.
    if debug_flag > 0:
        print(f"\nCalculating additional memory allocation requirements...")

    for ii in range(n_imgs+ONE_UINT32):
        for rr in range(n_rows+ONE_UINT32):
            for cc in range(n_cols+ONE_UINT32):
                vert_ids[ii,rr,cc] = MAX_UINT32

    # First, loop through the image sequence to find out how many vertices and
    # cells will actually be needed. This triple for-loop is just to allocate 
    # the vertex and cell arrays to the correct size later.
    for ii in range(n_imgs):
        for rr in range(n_rows):
            for cc in range(n_cols):

                temp_UINT8 = img_arr[ii,rr,cc]
                if temp_UINT8 < ONE_UINT8:
                    continue # Skip black voxels

                num_cells = num_cells + ONE_UINT32

                # Node 1
                if vert_ids[ii, rr, cc] == MAX_UINT32:
                    vert_ids[ii, rr, cc] = num_verts
                    num_verts = num_verts + ONE_UINT32

                # Node 2
                if vert_ids[ii, rr, cc+ONE_PYSZT] == MAX_UINT32:
                    vert_ids[ii, rr, cc+ONE_PYSZT] = num_verts
                    num_verts = num_verts + ONE_UINT32

                # Node 3
                if vert_ids[ii, rr+ONE_PYSZT, cc+ONE_PYSZT] == MAX_UINT32:
                    vert_ids[ii, rr+ONE_PYSZT, cc+ONE_PYSZT] = num_verts
                    num_verts = num_verts + ONE_UINT32

                # Node 4
                if vert_ids[ii, rr+ONE_PYSZT, cc] == MAX_UINT32:
                    vert_ids[ii, rr+ONE_PYSZT, cc] = num_verts
                    num_verts = num_verts + ONE_UINT32

                # Node 5
                if vert_ids[ii+ONE_PYSZT, rr, cc] == MAX_UINT32:
                    vert_ids[ii+ONE_PYSZT, rr, cc] = num_verts
                    num_verts = num_verts + ONE_UINT32

                # Node 6
                if vert_ids[ii+ONE_PYSZT, rr, cc+ONE_PYSZT] == MAX_UINT32:
                    vert_ids[ii+ONE_PYSZT, rr, cc+ONE_PYSZT] = num_verts
                    num_verts = num_verts + ONE_UINT32

                # Node 7
                if vert_ids[ii+ONE_PYSZT, rr+ONE_PYSZT, cc+ONE_PYSZT] == MAX_UINT32:
                    vert_ids[ii+ONE_PYSZT, rr+ONE_PYSZT, cc+ONE_PYSZT] = num_verts
                    num_verts = num_verts + ONE_UINT32

                # Node 8
                if vert_ids[ii+ONE_PYSZT, rr+ONE_PYSZT, cc] == MAX_UINT32:
                    vert_ids[ii+ONE_PYSZT, rr+ONE_PYSZT, cc] = num_verts
                    num_verts = num_verts + ONE_UINT32

    print(f"  Found {num_verts} vertices")
    print(f"  Found {num_cells} voxels")

    if debug_flag > 0:
        print(f"\nAllocating approximately "\
            f"{(num_verts*3 + num_cells*9)*4/1E6 + num_cells/1E6}"\
            f" MB of additional memory")

    cdef np.ndarray[DTYPE3_t, ndim=2] nd_coords = np.zeros(
        (num_verts,3), dtype=DTYPE3)

    cdef np.ndarray[DTYPE2_t, ndim=2] cell_conn = np.zeros(
        (num_cells,9), dtype=DTYPE2)

    cdef np.ndarray[DTYPE1_t, ndim=1] cell_vals = np.zeros(
        num_cells, dtype=DTYPE1)

    if (num_verts == 0) or (num_cells == 0):
        return [nd_coords, cell_conn, cell_vals]

    # Now that everything is allocated, re-initialize the vertex array
    print(f"\nBuilding voxel connectivity matrix...")
    for ii in range(n_imgs+ONE_UINT32):
        for rr in range(n_rows+ONE_UINT32):
            for cc in range(n_cols+ONE_UINT32):
                vert_ids[ii,rr,cc] = MAX_UINT32

    # This triple for-loop is the one that will actually store the nodal
    # coordinates of each vertex, and construct the connectivity matrix.
    # That is, the nodal indices (or Node IDs) used to define each cell 
    # (or voxel).
    num_verts = 0 # Also using these as indices, not just counters
    num_cells = 0 # Also using these as indices, not just counters
    for ii in range(n_imgs):
        for rr in range(n_rows):
            for cc in range(n_cols):

                temp_UINT8 = img_arr[ii,rr,cc]
                if temp_UINT8 < ONE_UINT8:
                    continue # Skip black voxels

                xc = cc # Switching the indices around!
                yc = rr
                zc = ii
                
                # Calculate vertex/node coordinates for each cell
                nd1[0] = xc - HALF
                nd1[1] = yc - HALF
                nd1[2] = zc - HALF
                
                nd2[0] = xc + HALF
                nd2[1] = yc - HALF
                nd2[2] = zc - HALF
                
                nd3[0] = xc + HALF
                nd3[1] = yc + HALF
                nd3[2] = zc - HALF
                
                nd4[0] = xc - HALF
                nd4[1] = yc + HALF
                nd4[2] = zc - HALF
                
                nd5[0] = xc - HALF
                nd5[1] = yc - HALF
                nd5[2] = zc + HALF
                
                nd6[0] = xc + HALF
                nd6[1] = yc - HALF
                nd6[2] = zc + HALF
                
                nd7[0] = xc + HALF
                nd7[1] = yc + HALF
                nd7[2] = zc + HALF
                
                nd8[0] = xc - HALF
                nd8[1] = yc + HALF
                nd8[2] = zc + HALF

                # Determine the node IDs for each cell. This requires
                # checking to see if the node already has been activated
                # in the vert_ids 3D array. If it has not, then activate 
                # it and store the new node coordinates into nd_coords.
                # If it has, then just need the node ID. Do this for each
                # node.

                # Node 1
                if vert_ids[ii, rr, cc] == MAX_UINT32:
                    vert_ids[ii, rr, cc] = num_verts 
                    nd1_id = num_verts # New node/vertex
                    nd_coords[num_verts,0] = nd1[0]
                    nd_coords[num_verts,1] = nd1[1]
                    nd_coords[num_verts,2] = nd1[2]
                    num_verts = num_verts + ONE_UINT32
                else:
                    # Existing node/vertex
                    nd1_id = vert_ids[ii, rr, cc] 

                # Node 2
                if vert_ids[ii, rr, cc+ONE_PYSZT] == MAX_UINT32:
                    vert_ids[ii, rr, cc+ONE_PYSZT] = num_verts 
                    nd2_id = num_verts # New node/vertex
                    nd_coords[num_verts,0] = nd2[0]
                    nd_coords[num_verts,1] = nd2[1]
                    nd_coords[num_verts,2] = nd2[2]
                    num_verts = num_verts + ONE_UINT32
                else:
                    # Existing node/vertex
                    nd2_id = vert_ids[ii, rr, cc+ONE_PYSZT]

                # Node 3
                if vert_ids[ii, rr+ONE_PYSZT, cc+ONE_PYSZT] == MAX_UINT32:
                    vert_ids[ii, rr+ONE_PYSZT, cc+ONE_PYSZT] = num_verts 
                    nd3_id = num_verts # New node/vertex
                    nd_coords[num_verts,0] = nd3[0]
                    nd_coords[num_verts,1] = nd3[1]
                    nd_coords[num_verts,2] = nd3[2]
                    num_verts = num_verts + ONE_UINT32
                else:
                    # Existing node/vertex
                    nd3_id = vert_ids[ii, rr+ONE_PYSZT, cc+ONE_PYSZT]

                # Node 4
                if vert_ids[ii, rr+ONE_PYSZT, cc] == MAX_UINT32:
                    vert_ids[ii, rr+ONE_PYSZT, cc] = num_verts
                    nd4_id = num_verts # New node/vertex
                    nd_coords[num_verts,0] = nd4[0]
                    nd_coords[num_verts,1] = nd4[1]
                    nd_coords[num_verts,2] = nd4[2]
                    num_verts = num_verts + ONE_UINT32
                else:
                    # Existing node/vertex
                    nd4_id = vert_ids[ii, rr+ONE_PYSZT, cc]

                # Node 5
                if vert_ids[ii+ONE_PYSZT, rr, cc] == MAX_UINT32:
                    vert_ids[ii+ONE_PYSZT, rr, cc] = num_verts
                    nd5_id = num_verts # New node/vertex
                    nd_coords[num_verts,0] = nd5[0]
                    nd_coords[num_verts,1] = nd5[1]
                    nd_coords[num_verts,2] = nd5[2]
                    num_verts = num_verts + ONE_UINT32
                else:
                    # Existing node/vertex
                    nd5_id = vert_ids[ii+ONE_PYSZT, rr, cc]

                # Node 6
                if vert_ids[ii+ONE_PYSZT, rr, cc+ONE_PYSZT] == MAX_UINT32:
                    vert_ids[ii+ONE_PYSZT, rr, cc+ONE_PYSZT] = num_verts
                    nd6_id = num_verts # New node/vertex
                    nd_coords[num_verts,0] = nd6[0]
                    nd_coords[num_verts,1] = nd6[1]
                    nd_coords[num_verts,2] = nd6[2]
                    num_verts = num_verts + ONE_UINT32
                else:
                    # Existing node/vertex
                    nd6_id = vert_ids[ii+ONE_PYSZT, rr, cc+ONE_PYSZT]

                # Node 7
                if vert_ids[ii+ONE_PYSZT, rr+ONE_PYSZT, cc+ONE_PYSZT] == MAX_UINT32:
                    vert_ids[ii+ONE_PYSZT, rr+ONE_PYSZT, cc+ONE_PYSZT] = num_verts
                    nd7_id = num_verts # New node/vertex
                    nd_coords[num_verts,0] = nd7[0]
                    nd_coords[num_verts,1] = nd7[1]
                    nd_coords[num_verts,2] = nd7[2]
                    num_verts = num_verts + ONE_UINT32
                else:
                    # Existing node/vertex
                    nd7_id = vert_ids[ii+ONE_PYSZT, rr+ONE_PYSZT, cc+ONE_PYSZT]

                # Node 8
                if vert_ids[ii+ONE_PYSZT, rr+ONE_PYSZT, cc] == MAX_UINT32:
                    vert_ids[ii+ONE_PYSZT, rr+ONE_PYSZT, cc] = num_verts
                    nd8_id = num_verts # New node/vertex
                    nd_coords[num_verts,0] = nd8[0]
                    nd_coords[num_verts,1] = nd8[1]
                    nd_coords[num_verts,2] = nd8[2]
                    num_verts = num_verts + ONE_UINT32
                else:
                    # Existing node/vertex
                    nd8_id = vert_ids[ii+ONE_PYSZT, rr+ONE_PYSZT, cc]

                # Store each node ID into the cell array. I call this the
                # connectivity matrix. Note, each row starts with the number
                # eight, then the node IDs are stored. This is a VTK thing,
                # and the number eight tells VTK to look for eight node IDs
                # and that this is a definition of a hexahedron.
                cell_conn[num_cells,0] = EIGHT
                cell_conn[num_cells,1] = nd1_id
                cell_conn[num_cells,2] = nd2_id
                cell_conn[num_cells,3] = nd3_id
                cell_conn[num_cells,4] = nd4_id
                cell_conn[num_cells,5] = nd5_id
                cell_conn[num_cells,6] = nd6_id
                cell_conn[num_cells,7] = nd7_id
                cell_conn[num_cells,8] = nd8_id

                # Store the intensity value for this voxel too
                cell_vals[num_cells] = temp_UINT8

                num_cells = num_cells + ONE_UINT32

    print(f"  Success!")
    return [nd_coords, cell_conn, cell_vals]


def extract_sphere(np.ndarray[DTYPE1_t, ndim=3] img_arr, DTYPE2_t radius):

    # Some basic checks for the input arrays
    cdef DTYPE2_t img_dim = img_arr.ndim

    if img_dim != 3:
        print(f"Number of dimensions in image array: {img_dim}")
        raise ValueError("Image array, img_arr, must be a 3D Numpy array")
    assert img_arr.dtype == DTYPE1

    # Py_ssize_t is essentially just an int (for the Python purist) which
    # should be used for index in Cython. Can actually use an int here 
    # and it would be fine
    cdef DTYPE2_t n_imgs = img_arr.shape[0]
    cdef DTYPE2_t n_rows = img_arr.shape[1]
    cdef DTYPE2_t n_cols = img_arr.shape[2]
    cdef DTYPE2_t n_i_half = (np.floor(n_imgs/2)).astype(DTYPE2)
    cdef DTYPE2_t n_r_half = (np.floor(n_rows/2)).astype(DTYPE2)
    cdef DTYPE2_t n_c_half = (np.floor(n_cols/2)).astype(DTYPE2)

    if (n_i_half-1) < radius:
        print(f"\nNumber of image slices in image array: {n_imgs}")
        print(f"Desired spherical radius: {radius}")
        print(f"WARNING: Radius is not less than half of"+\
            " the number of image slices in image array")
        print(f"Setting spherical radius to {n_i_half-1}\n")
        radius = n_i_half-1

    if (n_r_half-1) < radius:
        print(f"\nNumber of rows in image array: {n_rows}")
        print(f"Desired spherical radius: {radius}")
        print(f"WARNING: Radius is not less than half of"+\
            " the number of rows in image array")
        print(f"Setting spherical radius to {n_r_half-1}\n")
        radius = n_r_half-1

    if (n_c_half-1) < radius:
        print(f"\nNumber of columns in image array: {n_cols}")
        print(f"Desired spherical radius: {radius}")
        print(f"WARNING: Radius is not less than half of"+\
            " the number of columns in image array")
        print(f"Setting spherical radius to {n_c_half-1}\n")
        radius = n_c_half-1

    cdef Py_ssize_t ii, rr, cc, x, y, z,
    cdef DTYPE2_t a, b, c

    # The extracted sphere of voxels will be saved in this image 
    # sequence and returned
    cdef np.ndarray[DTYPE1_t, ndim=3] img_sph = np.zeros(
        (n_imgs, n_rows, n_cols), dtype=DTYPE1)

    cdef DTYPE5_t r_sqrd, lhs

    a = n_c_half # X center point
    b = n_r_half # Y center point
    c = n_i_half # Z center point
    r_sqrd = radius*radius

    # Loop through image sequence to find all of the voxels within a 
    # spherical radius from the center coordinate
    for ii in range(n_imgs):
        for rr in range(n_rows):
            for cc in range(n_cols): 

                x = cc
                y = rr
                z = ii

                # Sphere equation
                # (x-a)^2 + (y-b)^2 + (z-c)^2 = r^2
                lhs = (x-a)*(x-a) + (y-b)*(y-b) + (z-c)*(z-c)

                if lhs <= r_sqrd:
                    img_sph[ii,rr,cc] = img_arr[ii,rr,cc]

    return img_sph




"""
Written by Newell Moser
NRC Postdoctoral Fellow
National Institute of Standards and Technology
December 2, 2024

Test 3D rotations of image stacks.
"""

import os, pathlib, glob
import numpy as np
from skimage.util import img_as_ubyte, img_as_float32
import imppy3d.volume_image_processing as vol

#import imppy3d.vtk_api as vapi
#import imppy3d.import_export as imex
#import pyvista as pv


def make_dummy_img_stack(vec_dir='z'):
    imgs = np.zeros((129,129,129), dtype=np.uint8)

    rad_maj = 31
    rad_min = 1
    mid_i = int(np.floor(129/2.0))
    
    if vec_dir=='z': # Along image coordinates
        imgs[mid_i-rad_maj:mid_i+rad_maj+1, mid_i-rad_min:mid_i+rad_min+1,
            mid_i-rad_min:mid_i+rad_min+1] = 255

    elif vec_dir=='y': # Along row coordinates
        imgs[mid_i-rad_min:mid_i+rad_min+1, mid_i-rad_maj:mid_i+rad_maj+1, 
            mid_i-rad_min:mid_i+rad_min+1] = 255

    elif vec_dir=='x': # Along column coordinates
        imgs[mid_i-rad_min:mid_i+rad_min+1, mid_i-rad_min:mid_i+rad_min+1,
            mid_i-rad_maj:mid_i+rad_maj+1] = 255

    return imgs


def rotate_about_x():
    test_name = "\nTEST: ROTATING BINARY IMAGE STACK ABOUT X-AXIS..."
    print(test_name)

    imgs_src = make_dummy_img_stack('z')
    imgs_ref = make_dummy_img_stack('y')

    x_dir = np.array([0,0,1])
    rot_axis = x_dir
    rot_angle = 90*np.pi/180.0

    try:
        imgs_calc = vol.img_rotate_3D(imgs_src, rot_axis, rot_angle, 
            binarize=True)
    except:
        print(f"\nERROR: Failed to rotate the image stack.")
        return [1, test_name + " ERROR"]

    TOL = 1.0E-2
    imgs_diff = np.absolute(img_as_float32(imgs_calc) - img_as_float32(imgs_ref))
    max_diff = np.amax(imgs_diff)

    if max_diff >= TOL:
        print(f"\nERROR: Rotated image stack does not match reference image stack")
        return [1, test_name + " ERROR"]

    return [0, test_name + " SUCCESS"] # (False) No errors


def rotate_about_y():
    test_name = "\nTEST: ROTATING BINARY IMAGE STACK ABOUT Y-AXIS..."
    print(test_name)

    imgs_src = make_dummy_img_stack('z')
    imgs_ref = make_dummy_img_stack('x')

    y_dir = np.array([0,1,0])
    rot_axis = y_dir
    rot_angle = 90*np.pi/180.0

    try:
        imgs_calc = vol.img_rotate_3D(imgs_src, rot_axis, rot_angle, 
            binarize=True)
    except:
        print(f"\nERROR: Failed to rotate the image stack.")
        return [1, test_name + " ERROR"]

    TOL = 1.0E-2
    imgs_diff = np.absolute(img_as_float32(imgs_calc) - img_as_float32(imgs_ref))
    max_diff = np.amax(imgs_diff)

    if max_diff >= TOL:
        print(f"\nERROR: Rotated image stack does not match reference image stack")
        return [1, test_name + " ERROR"]

    return [0, test_name + " SUCCESS"] # (False) No errors


def rotate_about_z():
    test_name = "\nTEST: ROTATING BINARY IMAGE STACK ABOUT Z-AXIS..."
    print(test_name)

    imgs_src = make_dummy_img_stack('x')
    imgs_ref = make_dummy_img_stack('y')

    z_dir = np.array([1,0,0])
    rot_axis = z_dir
    rot_angle = 90*np.pi/180.0

    try:
        imgs_calc = vol.img_rotate_3D(imgs_src, rot_axis, rot_angle, 
            binarize=True)
    except:
        print(f"\nERROR: Failed to rotate the image stack.")
        return [1, test_name + " ERROR"]

    TOL = 1.0E-2
    imgs_diff = np.absolute(img_as_float32(imgs_calc) - img_as_float32(imgs_ref))
    max_diff = np.amax(imgs_diff)

    if max_diff >= TOL:
        print(f"\nERROR: Rotated image stack does not match reference image stack")
        return [1, test_name + " ERROR"]

    return [0, test_name + " SUCCESS"] # (False) No errors


def rotate_about_xyz():

    test_name = "\nTEST: ARBITRARY ROTATION OF BINARY IMAGE STACK..."
    print(test_name)

    imgs_src = make_dummy_img_stack('z')

    desired_vec3 = np.array([1.0,1.0,1.0])
    z_vec3 =  np.array([1.0,0.0,0.0])

    desired_vec3_mag = np.sqrt( np.power(desired_vec3[0], 2) + 
                            np.power(desired_vec3[1], 2) +
                            np.power(desired_vec3[2], 2) )
    desired_vec3 = desired_vec3/desired_vec3_mag

    rot_axis = np.cross(desired_vec3, z_vec3)
    rot_axis_mag = np.sqrt( np.power(rot_axis[0], 2) + 
                            np.power(rot_axis[1], 2) +
                            np.power(rot_axis[2], 2) )
    rot_axis = rot_axis/rot_axis_mag

    rot_angle = np.arccos(np.clip(np.dot(z_vec3, desired_vec3), -1.0, 1.0))

    # Apply the affine transformation
    print("\nPerforming rotation affine transformation...")
    print(f"    Rotation Axis (img, row, col): {rot_axis}")
    print(f"    Rotation Magnitude (CCW): {rot_angle*180.0/np.pi} (degrees)")

    try:
        imgs_calc = vol.img_rotate_3D(imgs_src, rot_axis, rot_angle, 
            binarize=True)
    except:
        print(f"\nERROR: Failed to rotate the image stack.")
        return [1, test_name + " ERROR"]

    xyz_pnts_tup = np.nonzero(imgs_calc)
    img_coords = np.asarray(xyz_pnts_tup[0])
    row_coords = np.asarray(xyz_pnts_tup[1])
    col_coords = np.asarray(xyz_pnts_tup[2])
    xyz_pnts_arr = np.column_stack((img_coords, row_coords, col_coords))

    # Fit a line to the newly rotated points
    try:
        mean_pnt, eig_vec1 = vol.line_fit_3d(xyz_pnts_arr)
    except:
        print(f"\nERROR: Could not fit a 3D line to the rotated "\
            "image stack.")
        return [1, test_name + " ERROR"]

    # Check within a tolerance of 3 degree
    angle_TOL = 3.0*np.pi/180.0 # Radians

    # The fitted line should be in the [1, 1, 1] direction
    paral_bools = np.isclose(desired_vec3, eig_vec1, rtol=0.0, atol=angle_TOL)
    if not all(paral_bools):
        print(f"\nERROR: Rotated image stack is misaligned.")
        return [1, test_name + " ERROR"]

    return [0, test_name + " SUCCESS"] # (False) No errors


if __name__ == '__main__':

    flag_200, msg_200 = rotate_about_x()
    flag_201, msg_201 = rotate_about_y()
    flag_202, msg_202 = rotate_about_z()
    flag_203, msg_203 = rotate_about_xyz()

    print(f"\n\n\n---------- SUMMARY ----------")
    print(msg_200)
    print(msg_201)
    print(msg_202)
    print(msg_203)
    print("\n")